@@ -7,7 +7,7 @@
  *
  * Description: Newt UI for cdebconf
  *
- * $Id: newt.c 42943 2006-11-27 10:50:31Z cjwatson $
+ * $Id: newt.c 57544 2009-02-13 16:44:52Z cjwatson $
  *
  * cdebconf is (c) 2000-2001 Randolph Chung and others under the following
  * license.
@@ -75,37 +75,43 @@ struct newt_data {
 };
 
 struct newtColors newtAltColorPalette = {
-	"white", "blue", 			/* root fg, bg */
-	/*"black", "lightgray",*/			/* border fg, bg */
+	/*"white", "blue",*/ 			/* root fg, bg */
+	"white", "black", 			/* root fg, bg */
+	/*"black", "lightgray",*/		/* border fg, bg */
 	"white", "black",			/* border fg, bg */
-	/*"black", "lightgray",*/			/* window fg, bg */
+	/*"black", "lightgray",*/		/* window fg, bg */
 	"white", "black",			/* window fg, bg */
 	/*"white", "black",*/			/* shadow fg, bg */
-	"white", "blue",			/* shadow fg, bg */
+	"white", "black",			/* shadow fg, bg */
 	/*"red", "lightgray",*/			/* title fg, bg */
 	"yellow", "black",			/* title fg, bg */
-	"brightred", "gray",			/* button fg, bg */
+	/*"brightred", "gray",*/		/* button fg, bg */
+	"black", "lightgray",			/* button fg, bg */
 	"gray", "brightred",			/* active button fg, bg */
 	"white", "gray",			/* checkbox fg, bg */
-	"yellow", "brown",			/* active checkbox fg, bg */
-	"yellow", "blue",			/* entry box fg, bg */
-	/*"blue", "lightgray",*/			/* label fg, bg */
+	/*"yellow", "brown",*/			/* active checkbox fg, bg */
+	"black", "lightgray",			/* active checkbox fg, bg */
+	/*"yellow", "blue",*/			/* entry box fg, bg */
+	"white", "black",			/* entry box fg, bg */
+	/*"blue", "lightgray",*/		/* label fg, bg */
 	"brightred", "black",			/* label fg, bg */
-	/*"black", "lightgray",*/			/* listbox fg, bg */
-	"lightgray", "black",			/* listbox fg, bg */
+	/*"black", "lightgray",*/		/* listbox fg, bg */
+	"white", "black",			/* listbox fg, bg */
 	"yellow", "blue",			/* active listbox fg, bg */
-	/*"black", "lightgray",*/			/* textbox fg, bg */
+	/*"black", "lightgray",*/		/* textbox fg, bg */
 	"white", "black",			/* textbox fg, bg */
 	"lightgray", "black",			/* active textbox fg, bg */
 	/*"white", "blue",*/			/* help line */
 	"white", "black",			/* help line */
 	"yellow", "blue",			/* root text */
-	"blue",					/* scale full */
-	"red",					/* scale empty */
+	/*"blue",*/				/* scale full */
+	"black",				/* scale full */
+	/*"red",*/				/* scale empty */
+	"blue",					/* scale empty */
 	"blue", "lightgray",			/* disabled entry fg, bg */
-	/*"black", "lightgray",*/			/* compact button fg, bg */
-	"lightgray", "black",			/* compact button fg, bg */
-	"yellow", "gray",			/* active & sel listbox */
+	/*"black", "lightgray",*/		/* compact button fg, bg */
+	"white", "black",			/* compact button fg, bg */
+	"black", "lightgray",			/* active & sel listbox */
 	"black", "brown"			/* selected listbox */
 };
 
@@ -113,16 +119,12 @@ typedef int (newt_handler)(struct frontend *obj, struct question *q);
 
 static void newt_progress_stop(struct frontend *obj);
 
-#include "config-newt.h"
+#include "cdebconf_newt.h"
 
-/*  Horizontal offset between buttons and text box */
-#define TEXT_PADDING 1
-/*  Horizontal offset between text box and borders */
-#define BUTTON_PADDING 4
 /*  Padding of title width, allows for leading "[!!] " before title
- *  and the two vertical bars that limit the border.
+ *  and the two vertical bars and spaces that limit the title.
  */
-#define TITLE_PADDING 8
+#define TITLE_PADDING 9
 
 /* gettext would be much nicer :-( */
 static const char *
@@ -162,7 +164,29 @@ help_text(struct frontend *obj)
 }
 
 void
-newt_create_window(const int width, const int height, const char *title, const char *priority)
+cdebconf_newt_setup(void)
+{
+    SLang_init_tty(0, 1, 0); /* disable flow control */
+    newtInit();
+    newtSetColors(newtAltColorPalette);
+    newtCls();
+}
+
+/* cdebconf-newt-terminal needs this in order to be able to restore the
+ * display properly after tearing down the terminal.
+ */
+char *
+cdebconf_newt_get_progress_info(struct frontend *obj)
+{
+    struct newt_data *data = (struct newt_data *)obj->data;
+    if (data->scale_info)
+        return strdup(data->scale_info);
+    else
+        return NULL;
+}
+
+void
+cdebconf_newt_create_window(const int width, const int height, const char *title, const char *priority)
 {
     static const char *sigils[][2] = {
         { "low",      "." },
@@ -192,7 +216,7 @@ newt_create_window(const int width, const int height, const char *title, const c
 }
 
 int
-newt_get_text_height(const char *text, int win_width)
+cdebconf_newt_get_text_height(const char *text, int win_width)
 {
     newtComponent textbox, f;
     int t_height;
@@ -220,14 +244,14 @@ newt_get_text_height(const char *text, int win_width)
 #endif
     t_height = newtTextboxGetNumLines(textbox);
     // This is needed so the textbox gets freed...ick
-    f = create_form(NULL);
+    f = cdebconf_newt_create_form(NULL);
     newtFormAddComponent(f, textbox);
     newtFormDestroy(f);
     return t_height;
 }
 
 int
-newt_get_text_width(const char *text)
+cdebconf_newt_get_text_width(const char *text)
 {
     int t_width = 0;
     const char *p = text;
@@ -249,15 +273,15 @@ newt_get_text_width(const char *text)
 }
 
 static int
-min_window_height(struct question *q, int win_width)
+min_window_height(struct frontend *obj, struct question *q, int win_width)
 {
     int height = 3;
     char *type = q->template->type;
     char *q_ext_text;
 
-    q_ext_text = q_get_extended_description(q);
+    q_ext_text = q_get_extended_description(obj, q);
     if (q_ext_text != NULL)
-        height = newt_get_text_height(q_ext_text, win_width) + 1;
+        height = cdebconf_newt_get_text_height(q_ext_text, win_width) + 1;
     if (strcmp(type, "multiselect") == 0 || strcmp(type, "select") == 0)
         height += 4; // at least three lines for choices + blank line
     else if (strcmp(type, "string") == 0 || strcmp(type, "password") == 0)
@@ -267,22 +291,22 @@ min_window_height(struct question *q, int win_width)
 }
 
 static int
-need_separate_window(struct question *q)
+need_separate_window(struct frontend *obj, struct question *q)
 {
     int width = 80, height = 24;
     int x;
 
     newtGetScreenSize(&width, &height);
-    x = min_window_height(q, width-7);
+    x = min_window_height(obj, q, width-7);
     return (x >= height-5);
 }
 
 static char *
-get_full_description(struct question *q)
+get_full_description(struct frontend *obj, struct question *q)
 {
     char *res = NULL;
-    char *descr = q_get_description(q);
-    char *ext_descr = q_get_extended_description(q);
+    char *descr = q_get_description(obj, q);
+    char *ext_descr = q_get_extended_description(obj, q);
 
     assert(descr);
     assert(ext_descr);
@@ -316,8 +340,8 @@ show_separate_window(struct frontend *obj, struct question *q)
 #else
     int flags = NEWT_FLAG_WRAP;
 #endif
-    char *descr = q_get_description(q);
-    char *ext_descr = q_get_extended_description(q);
+    char *descr = q_get_description(obj, q);
+    char *ext_descr = q_get_extended_description(obj, q);
 
     assert(descr);
     assert(ext_descr);
@@ -329,13 +353,13 @@ show_separate_window(struct frontend *obj, struct question *q)
         full_description = strdup(ext_descr);
     }
     else
-        full_description = get_full_description(q);
+        full_description = get_full_description(obj, q);
 
     newtGetScreenSize(&width, &height);
     win_width = width-7;
     /*  There are 5 characters for sigils, plus 4 for borders */
     strtruncate(obj->title, win_width-9);
-    t_height = newt_get_text_height(full_description, win_width);
+    t_height = cdebconf_newt_get_text_height(full_description, win_width);
     if (t_height+extra <= height-5)
         win_height = t_height+extra;
     else {
@@ -351,23 +375,23 @@ show_separate_window(struct frontend *obj, struct question *q)
     free(full_description);
     full_description = wrappedtext;
 #endif
-    t_width = newt_get_text_width(full_description);
-    t_width_descr = newt_get_text_width(descr);
+    t_width = cdebconf_newt_get_text_width(full_description);
+    t_width_descr = cdebconf_newt_get_text_width(descr);
     if (t_width_descr > t_width)
         t_width = t_width_descr;
-    t_width_buttons = 2*BUTTON_PADDING + newt_get_text_width(continue_text(obj)) + 2;
+    t_width_buttons = 2*BUTTON_PADDING + cdebconf_newt_get_text_width(continue_text(obj)) + 2;
     if (obj->methods.can_go_back(obj, q))
         //  Add an interspace
-        t_width_buttons += newt_get_text_width(goback_text(obj)) + 3;
+        t_width_buttons += cdebconf_newt_get_text_width(goback_text(obj)) + 3;
     if (t_width_buttons > t_width)
         t_width = t_width_buttons;
     if (win_width > t_width + 2*TEXT_PADDING + t_width_scroll)
         win_width = t_width + 2*TEXT_PADDING + t_width_scroll;
-    t_width_title = newt_get_text_width(obj->title) + TITLE_PADDING;
+    t_width_title = cdebconf_newt_get_text_width(obj->title) + TITLE_PADDING;
     if (t_width_title > win_width)
         win_width = t_width_title;
-    newt_create_window(win_width, win_height, obj->title, q->priority);
-    form = create_form(NULL);
+    cdebconf_newt_create_window(win_width, win_height, obj->title, q->priority);
+    form = cdebconf_newt_create_form(NULL);
     if (format_note)
         newtFormAddComponent(form, newtLabel((win_width - strwidth(descr))/2, 0, descr));
     textbox = newtTextbox(TEXT_PADDING, 1, t_width, t_height, flags);
@@ -414,7 +438,7 @@ generic_handler_string(struct frontend *obj, struct question *q, int eflags)
 #endif
     char *defval;
     const char *result;
-    char *full_description = get_full_description(q);
+    char *full_description = get_full_description(obj, q);
 
     eflags |= NEWT_ENTRY_SCROLL | NEWT_FLAG_RETURNEXIT;
     newtGetScreenSize(&width, &height);
@@ -428,7 +452,7 @@ generic_handler_string(struct frontend *obj, struct question *q, int eflags)
     full_description = wrappedtext;
 #endif
     if (full_description != NULL)
-        t_height = newt_get_text_height(full_description, win_width);
+        t_height = cdebconf_newt_get_text_height(full_description, win_width);
     else
         t_height = 0;
     if (t_height + 6 <= height-5)
@@ -439,20 +463,20 @@ generic_handler_string(struct frontend *obj, struct question *q, int eflags)
         t_width_scroll = 2;
     }
     t_height = win_height - 6;
-    t_width = newt_get_text_width(full_description);
-    t_width_buttons = 2*BUTTON_PADDING + newt_get_text_width(continue_text(obj)) + 2;
+    t_width = cdebconf_newt_get_text_width(full_description);
+    t_width_buttons = 2*BUTTON_PADDING + cdebconf_newt_get_text_width(continue_text(obj)) + 2;
     if (obj->methods.can_go_back(obj, q))
         //  Add an interspace
-        t_width_buttons += newt_get_text_width(goback_text(obj)) + 3;
+        t_width_buttons += cdebconf_newt_get_text_width(goback_text(obj)) + 3;
     if (t_width_buttons > t_width)
         t_width = t_width_buttons;
     if (win_width > t_width + 2*TEXT_PADDING + t_width_scroll)
         win_width = t_width + 2*TEXT_PADDING + t_width_scroll;
-    t_width_title = newt_get_text_width(obj->title) + TITLE_PADDING;
+    t_width_title = cdebconf_newt_get_text_width(obj->title) + TITLE_PADDING;
     if (t_width_title > win_width)
         win_width = t_width_title;
-    newt_create_window(win_width, win_height, obj->title, q->priority);
-    form = create_form(NULL);
+    cdebconf_newt_create_window(win_width, win_height, obj->title, q->priority);
+    form = cdebconf_newt_create_form(NULL);
     textbox = newtTextbox(TEXT_PADDING, 1, t_width, t_height, tflags);
     assert(textbox);
     if (full_description != NULL)
@@ -500,8 +524,8 @@ show_multiselect_window(struct frontend *obj, struct question *q, int show_ext_d
     size_t res;
     wchar_t c;
     int *tindex = NULL;
-    const char *indices = q_get_indices(q);
-    char *full_description = get_full_description(q);
+    const char *indices = q_get_indices(obj, q);
+    char *full_description = get_full_description(obj, q);
 #ifdef HAVE_LIBTEXTWRAP
     textwrap_t tw;
     char *wrappedtext;
@@ -513,20 +537,24 @@ show_multiselect_window(struct frontend *obj, struct question *q, int show_ext_d
     newtGetScreenSize(&width, &height);
     win_width = width-7;
     strtruncate(obj->title, win_width-9);
-    count = strgetargc(q_get_choices_vals(q));
+    count = strgetargc(q_get_choices_vals(obj, q));
     if (count <= 0)
         return DC_NOTOK;
     choices = malloc(sizeof(char *) * count);
     choices_trans = malloc(sizeof(char *) * count);
     tindex = malloc(sizeof(int) * count);
-    if (strchoicesplitsort(q_get_choices_vals(q), q_get_choices(q), indices, choices, choices_trans, tindex, count) != count)
+    if (strchoicesplitsort(q_get_choices_vals(obj, q), q_get_choices(obj, q), indices, choices, choices_trans, tindex, count) != count)
         return DC_NOTOK;
 
+    if (obj->methods.can_align(obj, q)) {
+        stralign(choices_trans, count);
+    }
+
     defvals = malloc(sizeof(char *) * count);
     defcount = strchoicesplit(question_getvalue(q, ""), defvals, count);
     answer = malloc(sizeof(char) * count);
     sel_height = count;
-    form = create_form(NULL);
+    form = cdebconf_newt_create_form(NULL);
 #ifdef HAVE_LIBTEXTWRAP
     textwrap_init(&tw);
     textwrap_columns(&tw, win_width - 2 - 2*TEXT_PADDING);
@@ -535,11 +563,11 @@ show_multiselect_window(struct frontend *obj, struct question *q, int show_ext_d
     full_description = wrappedtext;
 #endif
     sel_width = strlongest(choices_trans, count);
-    t_width = newt_get_text_width(full_description);
-    t_width_buttons = 2*BUTTON_PADDING + newt_get_text_width(continue_text(obj)) + 2;
+    t_width = cdebconf_newt_get_text_width(full_description);
+    t_width_buttons = 2*BUTTON_PADDING + cdebconf_newt_get_text_width(continue_text(obj)) + 2;
     if (obj->methods.can_go_back(obj, q))
         //  Add an interspace
-        t_width_buttons += newt_get_text_width(goback_text(obj)) + 3;
+        t_width_buttons += cdebconf_newt_get_text_width(goback_text(obj)) + 3;
     if (t_width_buttons > t_width)
         t_width = t_width_buttons;
     //  Truncate too wide items
@@ -560,7 +588,7 @@ show_multiselect_window(struct frontend *obj, struct question *q, int show_ext_d
         t_width = sel_width;
     if (win_width > t_width + 8)
         win_width = t_width + 8;
-    t_width_title = newt_get_text_width(obj->title) + TITLE_PADDING;
+    t_width_title = cdebconf_newt_get_text_width(obj->title) + TITLE_PADDING;
     if (t_width_title > win_width)
         win_width = t_width_title;
     if (show_ext_desc && full_description) {
@@ -578,7 +606,7 @@ show_multiselect_window(struct frontend *obj, struct question *q, int show_ext_d
         win_height = height-5;
         sel_height = win_height - t_height - 5;
     }
-    newt_create_window(win_width, win_height, obj->title, q->priority);
+    cdebconf_newt_create_window(win_width, win_height, obj->title, q->priority);
     if (count > sel_height) {
         scrollbar = newtVerticalScrollbar((win_width+sel_width+5)/2, 1+t_height+1, sel_height,
                 NEWT_COLORSET_WINDOW, NEWT_COLORSET_ACTCHECKBOX);
@@ -586,7 +614,7 @@ show_multiselect_window(struct frontend *obj, struct question *q, int show_ext_d
     }
     else
         scrollbar = NULL;
-    sform = create_form(scrollbar);
+    sform = cdebconf_newt_create_form(scrollbar);
     newtFormSetBackground(sform, NEWT_COLORSET_CHECKBOX);
     newtFormSetHeight(sform, sel_height);
     newtFormSetWidth(sform, sel_width+5);
@@ -652,8 +680,8 @@ show_select_window(struct frontend *obj, struct question *q, int show_ext_desc)
     char **choices, **choices_trans, *defval;
     int count = 0, i, ret, defchoice = -1;
     int *tindex = NULL;
-    const char *indices = q_get_indices(q);
-    char *full_description = get_full_description(q);
+    const char *indices = q_get_indices(obj, q);
+    char *full_description = get_full_description(obj, q);
     char *p;
     size_t res;
     int k;
@@ -669,17 +697,21 @@ show_select_window(struct frontend *obj, struct question *q, int show_ext_desc)
     newtGetScreenSize(&width, &height);
     win_width = width-7;
     strtruncate(obj->title, win_width-9);
-    count = strgetargc(q_get_choices_vals(q));
+    count = strgetargc(q_get_choices_vals(obj, q));
     if (count <= 0)
         return DC_NOTOK;
     choices = malloc(sizeof(char *) * count);
     choices_trans = malloc(sizeof(char *) * count);
     tindex = malloc(sizeof(int) * count);
-    if (strchoicesplitsort(q_get_choices_vals(q), q_get_choices(q), indices, choices, choices_trans, tindex, count) != count)
+    if (strchoicesplitsort(q_get_choices_vals(obj, q), q_get_choices(obj, q), indices, choices, choices_trans, tindex, count) != count)
         return DC_NOTOK;
 
+    if (obj->methods.can_align(obj, q)) {
+        stralign(choices_trans, count);
+    }
+
     sel_height = count;
-    form = create_form(NULL);
+    form = cdebconf_newt_create_form(NULL);
 #ifdef HAVE_LIBTEXTWRAP
     textwrap_init(&tw);
     textwrap_columns(&tw, win_width - 2 - 2*TEXT_PADDING);
@@ -688,11 +720,11 @@ show_select_window(struct frontend *obj, struct question *q, int show_ext_desc)
     full_description = wrappedtext;
 #endif
     sel_width = strlongest(choices_trans, count);
-    t_width = newt_get_text_width(full_description);
+    t_width = cdebconf_newt_get_text_width(full_description);
     t_width_buttons = 2*BUTTON_PADDING;
     if (obj->methods.can_go_back(obj, q))
         //  Add an interspace
-        t_width_buttons += newt_get_text_width(goback_text(obj)) + 3;
+        t_width_buttons += cdebconf_newt_get_text_width(goback_text(obj)) + 3;
     if (t_width_buttons > t_width)
         t_width = t_width_buttons;
     //  Truncate too wide items
@@ -713,7 +745,7 @@ show_select_window(struct frontend *obj, struct question *q, int show_ext_desc)
         t_width = sel_width;
     if (win_width > t_width + 8)
         win_width = t_width + 8;
-    t_width_title = newt_get_text_width(obj->title) + TITLE_PADDING;
+    t_width_title = cdebconf_newt_get_text_width(obj->title) + TITLE_PADDING;
     if (t_width_title > win_width)
         win_width = t_width_title;
     if (show_ext_desc && full_description) {
@@ -733,7 +765,7 @@ show_select_window(struct frontend *obj, struct question *q, int show_ext_desc)
     }
     if (count > sel_height)
         listflags |= NEWT_FLAG_SCROLL;
-    newt_create_window(win_width, win_height, obj->title, q->priority);
+    cdebconf_newt_create_window(win_width, win_height, obj->title, q->priority);
     listbox = newtListbox((win_width-sel_width-3)/2, 1+t_height+1, sel_height, listflags);
     defval = (char *)question_getvalue(q, "");
     for (i = 0; i < count; i++) {
@@ -792,7 +824,7 @@ newt_handler_boolean(struct frontend *obj, struct question *q)
 #else
     int flags = NEWT_FLAG_WRAP;
 #endif
-    char *full_description = get_full_description(q);
+    char *full_description = get_full_description(obj, q);
 
     newtGetScreenSize(&width, &height);
     win_width = width-7;
@@ -805,7 +837,7 @@ newt_handler_boolean(struct frontend *obj, struct question *q)
     full_description = wrappedtext;
 #endif
     if (full_description != NULL)
-        t_height = newt_get_text_height(full_description, win_width);
+        t_height = cdebconf_newt_get_text_height(full_description, win_width);
     else
         t_height = 0;
     if (t_height + 4 <= height-5)
@@ -816,23 +848,23 @@ newt_handler_boolean(struct frontend *obj, struct question *q)
         t_width_scroll = 2;
     }
     t_height = win_height - 4;
-    t_width = newt_get_text_width(full_description);
+    t_width = cdebconf_newt_get_text_width(full_description);
     t_width_buttons = 2*BUTTON_PADDING;
-    t_width_buttons += newt_get_text_width(yes_text(obj)) + 3;
-    t_width_buttons += newt_get_text_width(no_text(obj)) + 3;
+    t_width_buttons += cdebconf_newt_get_text_width(yes_text(obj)) + 3;
+    t_width_buttons += cdebconf_newt_get_text_width(no_text(obj)) + 3;
     if (obj->methods.can_go_back(obj, q))
         t_width_buttons += 3;
     if (obj->methods.can_go_back(obj, q))
-        t_width_buttons += newt_get_text_width(goback_text(obj)) + 3;
+        t_width_buttons += cdebconf_newt_get_text_width(goback_text(obj)) + 3;
     if (t_width_buttons > t_width)
         t_width = t_width_buttons;
     if (win_width > t_width + 2*TEXT_PADDING + t_width_scroll)
         win_width = t_width + 2*TEXT_PADDING + t_width_scroll;
-    t_width_title = newt_get_text_width(obj->title) + TITLE_PADDING;
+    t_width_title = cdebconf_newt_get_text_width(obj->title) + TITLE_PADDING;
     if (t_width_title > win_width)
         win_width = t_width_title;
-    newt_create_window(win_width, win_height, obj->title, q->priority);
-    form = create_form(NULL);
+    cdebconf_newt_create_window(win_width, win_height, obj->title, q->priority);
+    form = cdebconf_newt_create_form(NULL);
     textbox = newtTextbox(TEXT_PADDING, 1, t_width, t_height, flags);
     assert(textbox);
     if (full_description != NULL)
@@ -882,7 +914,7 @@ newt_handler_multiselect(struct frontend *obj, struct question *q)
 {
     int separate_window, ret;
 
-    separate_window = need_separate_window(q);
+    separate_window = need_separate_window(obj, q);
     while (1) {
         if (separate_window) {
             ret = show_separate_window(obj, q);
@@ -910,7 +942,7 @@ newt_handler_select(struct frontend *obj, struct question *q)
 {
     int separate_window, ret;
 
-    separate_window = need_separate_window(q);
+    separate_window = need_separate_window(obj, q);
     while (1) {
         if (separate_window) {
             ret = show_separate_window(obj, q);
@@ -984,12 +1016,14 @@ newt_handler_text(struct frontend *obj, struct question *q)
 static int
 newt_handler_error(struct frontend *obj, struct question *q)
 {
-    char *oldrootBg, *oldshadowBg;
+    char *oldrootBg, *oldshadowBg, *oldhelpLineFg, *oldhelpLineBg;
     int ret;
     struct newtColors palette = newtAltColorPalette;
 
     oldrootBg = palette.rootBg;
     oldshadowBg = palette.shadowBg;
+    oldhelpLineFg = palette.helpLineFg;
+    oldhelpLineBg = palette.helpLineBg;
     if (strcmp(oldrootBg, oldshadowBg) == 0)
     	palette.shadowBg = "red";
     palette.rootBg = "red";
@@ -999,6 +1033,8 @@ newt_handler_error(struct frontend *obj, struct question *q)
     ret = newt_handler_note(obj, q);
     palette.rootBg = oldrootBg;
     palette.shadowBg = oldshadowBg;
+    palette.helpLineFg = oldhelpLineFg;
+    palette.helpLineBg = oldhelpLineBg;
     newtSetColors(palette);
     return ret;
 }
@@ -1089,13 +1125,10 @@ newt_go(struct frontend *obj)
             if (plugin || strcmp(q->template->type, question_handlers[i].type) == 0) {
                 if (!cleared && !data->scale_form) {
                     cleared = 1;
-                    SLang_init_tty(0, 1, 0); /* disable flow control */
-                    newtInit();
-                    newtSetColors(newtAltColorPalette);
-                    newtCls();
+                    cdebconf_newt_setup();
                 }
                 if (obj->info != NULL) {
-                    char *text = q_get_description(obj->info);
+                    char *text = q_get_description(obj, obj->info);
                     if (text)
                         newtDrawRootText(0, 0, text);
                     free(text);
@@ -1105,26 +1138,24 @@ newt_go(struct frontend *obj)
                 newtPopHelpLine();
                 if (ret == DC_OK)
                     obj->qdb->methods.set(obj->qdb, q);
-                else if (ret == DC_GOBACK && q->prev != NULL)
-                    q = q->prev;
-                else {
-                    if (cleared && !data->scale_form)
-                        newtFinished();
-                    if (plugin)
-                        plugin_delete(plugin);
-                    return ret;
-                }
                 if (plugin)
                     plugin_delete(plugin);
                 break;
             }
         }
+        if (ret == DC_NOTOK)
+            break;
         if (ret == DC_OK)
             q = q->next;
+        else if (ret == DC_GOBACK) {
+            do {
+                q = q->prev;
+            } while (q != NULL && 0 == strcmp("error", q->template->type));
+	}
     }
     if (cleared && !data->scale_form)
         newtFinished();
-    return DC_OK;
+    return ret;
 }
 
 static bool
@@ -1139,6 +1170,28 @@ newt_can_cancel_progress(struct frontend *obj)
     return (obj->capability & DCF_CAPB_PROGRESSCANCEL);
 }
 
+static bool
+newt_can_align(struct frontend *obj, struct question *q)
+{
+    return (obj->capability & DCF_CAPB_ALIGN);
+}
+
+static const char *
+newt_lookup_directive(struct frontend *obj, const char *directive)
+{
+    if (obj->methods.can_align(obj, obj->questions)) {
+        if (strcmp("TAB", directive) == 0) {
+            return STRALIGN_TAB;
+        } else if (strcmp("ALIGN=CENTER", directive) == 0) {
+            return STRALIGN_ALIGN_CENTER;
+        } else if (strcmp("ALIGN=RIGHT", directive) == 0) {
+            return STRALIGN_ALIGN_RIGHT;
+        }
+    }
+    /* Remove unhandled directives */
+    return "";
+}
+
 static void
 newt_make_progress_bar(struct frontend *obj, const char *info)
 {
@@ -1165,7 +1218,7 @@ newt_make_progress_bar(struct frontend *obj, const char *info)
     else
         data->scale_info = strdup(info);
     if (info)
-        text_height = newt_get_text_height(info, win_width);
+        text_height = cdebconf_newt_get_text_height(info, win_width);
     else
         text_height = 0;
     /* Minimal height set to 2 to prevent box flashing */
@@ -1192,7 +1245,7 @@ newt_make_progress_bar(struct frontend *obj, const char *info)
         data->scale_bar = newtScale(TEXT_PADDING, 1, win_width-2*TEXT_PADDING, obj->progress_max - obj->progress_min);
         data->scale_textbox = newtTextbox(TEXT_PADDING, 3, win_width-2*TEXT_PADDING, text_height, flags);
         data->scale_textbox_height = text_height;
-        data->scale_form = create_form(NULL);
+        data->scale_form = cdebconf_newt_create_form(NULL);
         newtFormAddComponents(data->scale_form, data->scale_bar, data->scale_textbox, NULL);
         if (obj->methods.can_cancel_progress(obj)) {
             data->scale_cancel = newtCompactButton(TEXT_PADDING + BUTTON_PADDING - 1, win_height-2, cancel_text(obj));
@@ -1231,12 +1284,9 @@ newt_progress_start(struct frontend *obj, int min, int max, const char *title)
     obj->progress_max = max;
     obj->progress_cur = min;
     data->scale_info = NULL;
-    SLang_init_tty(0, 1, 0); /* disable flow control */
-    newtInit();
-    newtSetColors(newtAltColorPalette);
-    newtCls();
+    cdebconf_newt_setup();
     if (obj->info != NULL) {
-        char *text = q_get_description(obj->info);
+        char *text = q_get_description(obj, obj->info);
         if (text)
             newtDrawRootText(0, 0, text);
         free(text);
@@ -1333,9 +1383,11 @@ struct frontend_module debconf_frontend_module =
 {
 initialize: newt_initialize,
 shutdown: newt_shutdown,
+lookup_directive: newt_lookup_directive,
 go: newt_go,
 can_go_back: newt_can_go_back,
 can_cancel_progress: newt_can_cancel_progress,
+can_align: newt_can_align,
 progress_start: newt_progress_start,
 progress_set:   newt_progress_set,
 progress_info:  newt_progress_info,
