@@ -10,7 +10,7 @@
  * friendly implementation. I've taken care to make the prompts work well
  * with screen readers and the like.
  *
- * $Id: text.c 34558 2006-01-27 01:35:19Z cjwatson $
+ * $Id: text.c 55712 2008-09-09 21:16:05Z lunar $
  *
  * cdebconf is (c) 2000-2001 Randolph Chung and others under the following
  * license.
@@ -78,30 +78,6 @@ typedef int (text_handler)(struct frontend *obj, struct question *q);
 #define ISEMPTY(buf) (buf[0] == 0)
 #endif
 
-/*  This function will be moved to strutl.c  */
-/*
- * Add spaces at the end of string so that strwidth(that) == maxsize
- * Input string must have been allocated with enough space.
- */
-int
-strpad (char *what, size_t maxsize)
-{
-    size_t pos;
-    int k;
-    char *p;
-    wchar_t c;
-
-    pos = 0;
-    for (p = what; (k = mbtowc (&c, p, MB_LEN_MAX)) > 0; p += k)
-        pos += wcwidth (c);
-    if (pos > maxsize)
-        return 0;
-    for (k = pos; k < maxsize; k++, p++)
-        *p = ' ';
-    *p = '\0';
-    return 1;
-}  
-
 /*
  * Function: getwidth
  * Input: none
@@ -161,8 +137,8 @@ static void wrap_print(const char *str)
  */
 static void text_handler_displaydesc(struct frontend *obj, struct question *q) 
 {
-	char *descr = q_get_description(q);
-	char *ext_descr = q_get_extended_description(q);
+	char *descr = q_get_description(obj, q);
+	char *ext_descr = q_get_extended_description(obj, q);
 	if (strcmp(q->template->type, "note") == 0 ||
 	    strcmp(q->template->type, "error") == 0)
 	{
@@ -194,7 +170,7 @@ get_answer(char *answer, int size)
 static void
 show_help (struct frontend *obj, struct question *q)
 {
-	char *descr = q_get_description(q);
+	char *descr = q_get_description(obj, q);
 	printf("%s", question_get_text(obj, "debconf/text-help-keystrokes", "KEYSTROKES:"));
 	printf("\n  %c  ", CHAR_HELP);
 	printf("%s", question_get_text(obj, "debconf/text-help-help", "Display this help message"));
@@ -232,12 +208,18 @@ printlist (struct frontend *obj, struct question *q, int count, char **choices_t
 		logcount++;
 	} while (i > 0);
 
+	if (obj->methods.can_align(obj, q)) {
+		stralign(choices_translated, count);
+		/* Display only one column to get meaningful alignment. */
+		width = 1;
+	}
+
 	/*  Set string arrays  */
 	for (i=0; i < count; i++)
 	{
 		/*  Trailing spaces are a placeholder to add [*] for
 		    selected values */
-		asprintf(&(fchoices[i]), "  %d. %s    ", i+1, choices_translated[i]);
+		asprintf(&(fchoices[i]), "%3d. %s    ", i+1, choices_translated[i]);
 		if (selected[tindex[i]])
 			strcpy(fchoices[i]+strlen(fchoices[i])-3, "[*]");
 		if (strwidth(fchoices[i]) < choice_min || choice_min == -1)
@@ -388,9 +370,9 @@ static int text_handler_multiselect(struct frontend *obj, struct question *q)
 	int i, j, count = 0, dcount, choice;
 	int *tindex = NULL;
 	int ret = DC_OK;
-	const char *indices = q_get_indices(q);
+	const char *indices = q_get_indices(obj, q);
 
-	count = strgetargc(q_get_choices_vals(q));
+	count = strgetargc(q_get_choices_vals(obj, q));
 	if (count <= 0)
 		return DC_NOTOK;
 	choices = malloc(sizeof(char *) * count);
@@ -400,7 +382,7 @@ static int text_handler_multiselect(struct frontend *obj, struct question *q)
 		choices[i] = choices_translated[i] = NULL;
 	}
 	tindex = malloc(sizeof(int) * count);
-	if (strchoicesplitsort(q_get_choices_vals(q), q_get_choices(q), indices, choices, choices_translated, tindex, count) != count)
+	if (strchoicesplitsort(q_get_choices_vals(obj, q), q_get_choices(obj, q), indices, choices, choices_translated, tindex, count) != count)
 	{
 		ret = DC_NOTOK;
 		goto CleanUp_TINDEX;
@@ -515,9 +497,9 @@ static int text_handler_select(struct frontend *obj, struct question *q)
 	const char *defval = question_getvalue(q, "");
 	int *tindex = NULL;
 	int ret = DC_OK;
-	const char *indices = q_get_indices(q);
+	const char *indices = q_get_indices(obj, q);
 
-	count = strgetargc(q_get_choices_vals(q));
+	count = strgetargc(q_get_choices_vals(obj, q));
 	if (count <= 0)
 		return DC_NOTOK;
 	choices = malloc(sizeof(char *) * count);
@@ -527,7 +509,7 @@ static int text_handler_select(struct frontend *obj, struct question *q)
 		choices[i] = choices_translated[i] = NULL;
 	}
 	tindex = malloc(sizeof(int) * count);
-	if (strchoicesplitsort(q_get_choices_vals(q), q_get_choices(q), indices, choices, choices_translated, tindex, count) != count)
+	if (strchoicesplitsort(q_get_choices_vals(obj, q), q_get_choices(obj, q), indices, choices, choices_translated, tindex, count) != count)
 	{
 		ret = DC_NOTOK;
 		goto CleanUp_TINDEX;
@@ -769,10 +751,18 @@ struct question_handlers {
 static int text_initialize(struct frontend *obj, struct configuration *conf)
 {
 	struct frontend_data *data = NEW(struct frontend_data);
+	char *term = getenv("TERM");
+	char *palette = getenv("FRONTEND_BACKGROUND");
 	data->previous_title = NULL;
 	obj->data = data;
 	obj->interactive = 1;
 	signal(SIGINT, SIG_IGN);
+	if (palette && !strcmp(palette, "dark") &&
+			term && (!strcmp(term, "linux") || !strcmp(term, "bterm"))) {
+		/* Hard-code for these cases */
+		printf("\e[37m\e[40m\e[1m\e[H\e[J");
+		fflush(stdout);
+	}
 	return DC_OK;
 }
 
@@ -790,6 +780,44 @@ text_can_go_back(struct frontend *obj, struct question *q)
 	return (obj->capability & DCF_CAPB_BACKUP);
 }
 
+/*
+ * Function: text_can_align
+ * Input: struct frontend *obj - frontend object
+ *        struct question *q - question object
+ * Output: int - DC_OK, DC_NOTOK
+ * Description: tells whether confmodule supports aligning columns
+ * Assumptions: none
+ */
+static bool
+text_can_align(struct frontend *obj, struct question *q)
+{
+	return (obj->capability & DCF_CAPB_ALIGN);
+}
+
+/*
+ * Function: text_lookup_directive
+ * Input: struct frontend *obj - frontend object
+ *        const char *directive - the directive
+ * Output: const char * - directive value or NULL
+ * Description: returns directive value in the given context
+ * Assumptions: none
+ */
+static const char *
+text_lookup_directive(struct frontend *obj, const char *directive)
+{
+	if (obj->methods.can_align(obj, obj->questions)) {
+		if (strcmp("TAB", directive) == 0) {
+			return STRALIGN_TAB;
+		} else if (strcmp("ALIGN=CENTER", directive) == 0) {
+			return STRALIGN_ALIGN_CENTER;
+		} else if (strcmp("ALIGN=RIGHT", directive) == 0) {
+			return STRALIGN_ALIGN_RIGHT;
+		}
+	}
+	/* Remove unhandled directives */
+	return "";
+}
+
 /*
  * Function: text_go
  * Input: struct frontend *obj - frontend object
@@ -852,24 +880,24 @@ static int text_go(struct frontend *obj)
 				putchar('\n');
 				if (ret == DC_OK)
 					obj->qdb->methods.set(obj->qdb, q);
-				else if (ret == DC_GOBACK && q->prev != NULL)
-					q = q->prev;
-				else {
-					if (plugin)
-						plugin_delete(plugin);
-					return ret;
-				}
 				if (plugin)
 					plugin_delete(plugin);
 				break;
 			}
 		}
+		if (ret == DC_NOTOK)
+			break;
 		if (i == DIM(question_handlers))
 			return DC_NOTIMPL;
 		if (ret == DC_OK)
 			q = q->next;
+		else if (ret == DC_GOBACK) {
+			do {
+			    q = q->prev;
+			} while (q != NULL && 0 == strcmp("error", q->template->type));
+		}
 	}
-	return DC_OK;
+	return ret;
 }
 
 static void text_progress_start(struct frontend *obj, int min, int max, const char *title)
@@ -915,7 +943,9 @@ static void text_progress_stop(struct frontend *obj)
 struct frontend_module debconf_frontend_module =
 {
 	initialize: text_initialize,
+	lookup_directive: text_lookup_directive,
 	can_go_back: text_can_go_back,
+	can_align: text_can_align,
 	go: text_go,
 	progress_start: text_progress_start,
 	progress_set: text_progress_set,
