@@ -4,15 +4,11 @@
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdbool.h>
 
-static char *template_lget(struct template *t, const char *lang,
-                const char *field);
-static char *template_get_internal(struct template *t, const char *lang,
-                const char *field);
-static void template_lset(struct template *t, const char *lang,
-                const char *field, const char *value);
-static char *template_next_lang(struct template *t, const char *l);
-static struct template_l10n_fields *template_cur_l10n_fields(struct template *t,
+static const char *template_get_internal(const struct template *t,
+                const char *lang, const char *field);
+static struct template_l10n_fields *template_cur_l10n_fields(const struct template *t,
                 const char *lang);
 static char *getlanguage(void);
 static void remove_newlines(char *);
@@ -86,6 +82,28 @@ static char *getlanguage(void)
 	return cache_list_lang_ptr->lang;
 }
 
+static bool allow_i18n(void)
+{
+	const char *i18nenv = getenv("DEBCONF_NO_I18N");
+	if (i18nenv && !strcmp(i18nenv, "1"))
+		return false;
+	else
+		return true;
+}
+
+bool load_all_translations(void)
+{
+	static int translations = -1;
+	if (translations == -1) {
+		const char *translations_env = getenv("DEBCONF_DROP_TRANSLATIONS");
+		if (translations_env && !strcmp(translations_env, "1"))
+			translations = 0;
+		else
+			translations = 1;
+	}
+	return (translations == 1);
+}
+
 /*
  * Function: template_new
  * Input: a tag, describing which template this is.  Can be null.
@@ -102,13 +120,10 @@ struct template *template_new(const char *tag)
 	struct template_l10n_fields *f = NEW(struct template_l10n_fields);
 	struct template *t = NEW(struct template);
 	memset(f, 0, sizeof(struct template_l10n_fields));
-	f->language = strdup("C");
+	f->language = strdup("");
 	memset(t, 0, sizeof(struct template));
 	t->ref = 1;
 	t->tag = STRDUP(tag);
-	t->lget = template_lget;
-	t->lset = template_lset;
-	t->next_lang = template_next_lang;
 	t->fields = f;
 	return t;
 }
@@ -124,6 +139,7 @@ void template_delete(struct template *t)
 	while (p != NULL)
 	{
 		q = p->next;
+		DELETE(p->language);
 		DELETE(p->defaultval);
 		DELETE(p->choices);
 		DELETE(p->indices);
@@ -153,7 +169,7 @@ void template_deref(struct template *t)
  * Assumptions: template_new succeeds, STRDUP succeeds.
  */
 
-struct template *template_dup(struct template *t)
+struct template *template_dup(const struct template *t)
 {
         struct template *ret = template_new(t->tag);
         struct template_l10n_fields *from, *to;
@@ -190,7 +206,7 @@ struct template *template_dup(struct template *t)
         return ret;
 }
 
-struct template *template_l10nmerge(struct template *ret, struct template *t)
+struct template *template_l10nmerge(struct template *ret, const struct template *t)
 {
         struct template_l10n_fields *from, *to;
         int same_choices, same_description;
@@ -258,7 +274,29 @@ struct template *template_l10nmerge(struct template *ret, struct template *t)
         return ret;
 }
 
-static char *template_field_get(struct template_l10n_fields *p,
+void template_l10nclear(struct template *t)
+{
+    struct template_l10n_fields *p = t->fields, *q;
+
+    while (p != NULL)
+    {
+        q = p->next;
+        DELETE(p->language);
+        DELETE(p->defaultval);
+        DELETE(p->choices);
+        DELETE(p->indices);
+        DELETE(p->description);
+        DELETE(p->extended_description);
+        DELETE(p);
+        p = q;
+    }
+
+    t->fields = NEW(struct template_l10n_fields);
+    memset(t->fields, 0, sizeof(struct template_l10n_fields));
+    t->fields->language = strdup("");
+}
+
+static const char *template_field_get(const struct template_l10n_fields *p,
                 const char *field)
 {
     if (strcasecmp(field, "default") == 0)
@@ -316,10 +354,10 @@ static void template_field_set(struct template_l10n_fields *p,
  * Assumptions: 
  */
 
-static char *template_lget(struct template *t, const char *lang,
-                const char *field)
+const char *template_lget(const struct template *t,
+                const char *lang, const char *field)
 {
-    char *ret = NULL;
+    const char *ret = NULL;
     char *orig_field;
     char *altlang;
     char *cp;
@@ -337,16 +375,24 @@ static char *template_lget(struct template *t, const char *lang,
         altlang = strchr(orig_field, '-');
         *altlang = 0;
         altlang++;
-        cp = strstr(altlang, ".UTF-8");
-        if (cp + 6 == altlang + strlen(altlang) && cp != altlang + 1)
-        {
-            *cp = 0;
-            ret = template_lget(t, altlang, orig_field);
-        }
+        if (strcasecmp(altlang, "C") == 0)
+            ret = template_lget(t, "C", orig_field);
+        else {
+            if (!allow_i18n()) {
+                free(orig_field);
+                return NULL;
+            }
+            cp = strcasestr(altlang, ".UTF-8");
+            if (cp + 6 == altlang + strlen(altlang) && cp != altlang + 1)
+            {
+                *cp = 0;
+                ret = template_lget(t, altlang, orig_field);
+            }
 #ifndef NODEBUG
-        else
-            fprintf(stderr, "Unknown localized field:\n%s\n", field);
+            else
+                fprintf(stderr, "Unknown localized field:\n%s\n", field);
 #endif
+        }
         free(orig_field);
         return ret;
     }
@@ -385,14 +431,14 @@ static char *template_lget(struct template *t, const char *lang,
  *         description and extended_description
  * Description: get field value
  * Assumptions: Arguments have been previously checked, lang and field
- *              are not MULL
+ *              are not NULL
  */
 
-static char *template_get_internal(struct template *t, const char *lang,
-                const char *field)
+static const char *template_get_internal(const struct template *t,
+                const char *lang, const char *field)
 {
-    struct template_l10n_fields *p;
-    char *altret = NULL;
+    const struct template_l10n_fields *p;
+    const char *altret = NULL;
 
     p = t->fields->next;
     while (p != NULL)
@@ -413,14 +459,14 @@ static char *template_get_internal(struct template *t, const char *lang,
     return NULL;
 }
 
-static void template_lset(struct template *t, const char *lang,
+void template_lset(struct template *t, const char *lang,
                 const char *field, const char *value)
 {
     struct template_l10n_fields *p, *last;
     char *orig_field;
     char *altlang;
     char *cp;
-    const char *curlang;
+    const char *curlang = NULL;
 
     if (strcasecmp(field, "tag") == 0)
     {
@@ -433,23 +479,44 @@ static void template_lset(struct template *t, const char *lang,
         return;
     }
 
-    /*   If field is Foo-xx.UTF-8 then call template_lget(t, "xx", "Foo")  */
+    /*   If field is Foo-xx.UTF-8 then call template_lset(t, "xx", "Foo")  */
     if (strchr(field, '-') != NULL)
     {
         orig_field = strdup(field);
         altlang = strchr(orig_field, '-');
         *altlang = 0;
         altlang++;
-        cp = strstr(altlang, ".UTF-8");
-        if (cp + 6 == altlang + strlen(altlang) && cp != altlang + 1)
-        {
-            *cp = 0;
-            template_lset(t, altlang, orig_field, value);
-        }
+        if (strcasecmp(altlang, "C") == 0)
+            template_lset(t, "C", orig_field, value);
+        else {
+            if (!allow_i18n()) {
+                free(orig_field);
+                return;
+            }
+            cp = strcasestr(altlang, ".UTF-8");
+
+            /* Plain debconf supports undefined character sets, on the
+               form "Description-nb_NO: ", which is valid if the text is
+               ASCII (but debconf still uses that syntax regardless of
+               validity if the application does not specify a character
+               set). To avoid losing these fields, we simply read them
+               in as if they were UTF-8 fields, as valid ASCII is always
+               valid UTF-8 as well.
+              
+               -- sesse, 2006-06-19
+            */
+            if ((cp + 6 == altlang + strlen(altlang) && cp != altlang + 1)
+	        || strchr(altlang, '.') == NULL)
+            {
+                if (cp != NULL)
+                    *cp = 0;
+                template_lset(t, altlang, orig_field, value);
+            }
 #ifndef NODEBUG
-        else
-            fprintf(stderr, "Unknown localized field:\n%s\n", field);
+            else
+                fprintf(stderr, "Unknown localized field:\n%s\n", field);
 #endif
+        }
         free(orig_field);
         return;
     }
@@ -459,8 +526,33 @@ static void template_lset(struct template *t, const char *lang,
 
     if (*lang == 0)
         curlang = getlanguage();
-    else
+    else if (load_all_translations() ||
+             strcmp(lang, "C") == 0 || strncmp(lang, "en", 2) == 0)
         curlang = lang;
+    else {
+        char *p;
+        struct cache_list_lang *cl;
+
+        getlanguage();
+        for (cl = cache_list_lang_ptr; cl != NULL; cl = cl->next) {
+            char *wantlang;
+
+            wantlang = strdup(cl->lang);
+            p = strpbrk(wantlang, "_.@");
+            if (p)
+                *p = '\0';
+            if (strncmp(lang, wantlang, strlen(wantlang)) == 0) {
+                curlang = lang;
+                free(wantlang);
+                break;
+            }
+            free(wantlang);
+        }
+        if (curlang == NULL) {
+            INFO(INFO_VERBOSE, "Dropping %s/%s for %s", t->tag, field, lang);
+            return;
+        }
+    }
 
     p = t->fields;
     last = p;
@@ -481,7 +573,7 @@ static void template_lset(struct template *t, const char *lang,
     template_field_set(p, field, value);
 }
 
-static struct template_l10n_fields *template_cur_l10n_fields(struct template *t,
+static struct template_l10n_fields *template_cur_l10n_fields(const struct template *t,
                 const char *lang)
 {
     struct template_l10n_fields *p;
@@ -496,9 +588,9 @@ static struct template_l10n_fields *template_cur_l10n_fields(struct template *t,
     return NULL;
 }
 
-static char *template_next_lang(struct template *t, const char *lang)
+const char *template_next_lang(const struct template *t, const char *lang)
 {
-    struct template_l10n_fields *p = template_cur_l10n_fields(t, lang);
+    const struct template_l10n_fields *p = template_cur_l10n_fields(t, lang);
     if (p != NULL && p->next != NULL)
         return p->next->language;
     return NULL;
@@ -555,7 +647,8 @@ struct template *template_load(const char *filename)
 	struct template *tlist = NULL, *t = 0;
 	unsigned int i;
 	int linesize;
-	
+	bool i18n = allow_i18n();
+
 	if ((fp = fopen(filename, "r")) == NULL)
 		return NULL;
 	while (fgets(buf, sizeof(buf), fp))
@@ -588,56 +681,71 @@ struct template *template_load(const char *filename)
 			template_lset(t, NULL, "type", p+6);
 		else if (strstr(p, "Default: ") == p && t != 0)
 			template_lset(t, NULL, "default", p+9);
-		else if (strstr(p, "Default-") == p && t != 0) 
+		else if (i18n && strstr(p, "Default-") == p && t != 0)
 		{
-			cp = strstr(p, ".UTF-8: ");
-			if (cp != NULL && cp != p+8)
-			{
-				lang = strndup(p+8, (int) (cp - p - 8));
-				template_lset(t, lang, "default", cp+8);
-			}
+			if (strcasestr(p, "Default-C: ") == p)
+				template_lset(t, "C", "default", p+11);
 			else
 			{
+				cp = strcasestr(p, ".UTF-8: ");
+				if (cp != NULL && cp != p+8)
+				{
+					lang = strndup(p+8, (int) (cp - p - 8));
+					template_lset(t, lang, "default", cp+8);
+				}
+				else
+				{
 #ifndef NODEBUG
-				fprintf(stderr, "Unknown localized field:\n%s\n", p);
+					fprintf(stderr, "Unknown localized field:\n%s\n", p);
 #endif
-                                continue;
+					continue;
+				}
 			}
 		}
 		else if (strstr(p, "Choices: ") == p && t != 0)
 			template_lset(t, NULL, "choices", p+9);
-		else if (strstr(p, "Choices-") == p && t != 0) 
+		else if (i18n && strstr(p, "Choices-") == p && t != 0)
 		{
-			cp = strstr(p, ".UTF-8: ");
-			if (cp != NULL && cp != p+8)
-			{
-				lang = strndup(p+8, (int) (cp - p - 8));
-				template_lset(t, lang, "choices", cp+8);
-			}
+			if (strcasestr(p, "Choices-C: ") == p)
+				template_lset(t, "C", "choices", p+11);
 			else
 			{
+				cp = strcasestr(p, ".UTF-8: ");
+				if (cp != NULL && cp != p+8)
+				{
+					lang = strndup(p+8, (int) (cp - p - 8));
+					template_lset(t, lang, "choices", cp+8);
+				}
+				else
+				{
 #ifndef NODEBUG
-				fprintf(stderr, "Unknown localized field:\n%s\n", p);
+					fprintf(stderr, "Unknown localized field:\n%s\n", p);
 #endif
-                                continue;
+					continue;
+				}
 			}
 		}
 		else if (strstr(p, "Indices: ") == p && t != 0)
 			template_lset(t, NULL, "indices", p+9);
-		else if (strstr(p, "Indices-") == p && t != 0) 
+		else if (i18n && strstr(p, "Indices-") == p && t != 0)
 		{
-			cp = strstr(p, ".UTF-8: ");
-			if (cp != NULL && cp != p+8)
-			{
-				lang = strndup(p+8, (int) (cp - p - 8));
-				template_lset(t, lang, "indices", cp+8);
-			}
+			if (strcasestr(p, "Indices-C: ") == p)
+				template_lset(t, "C", "indices", p+11);
 			else
 			{
+				cp = strcasestr(p, ".UTF-8: ");
+				if (cp != NULL && cp != p+8)
+				{
+					lang = strndup(p+8, (int) (cp - p - 8));
+					template_lset(t, lang, "indices", cp+8);
+				}
+				else
+				{
 #ifndef NODEBUG
-				fprintf(stderr, "Unknown localized field:\n%s\n", p);
+					fprintf(stderr, "Unknown localized field:\n%s\n", p);
 #endif
-                                continue;
+					continue;
+				}
 			}
 		}
 		else if (strstr(p, "Description: ") == p && t != 0)
@@ -656,7 +764,8 @@ struct template *template_load(const char *filename)
 			{
                                 if (i != ungetc(i, fp)) { /* ERROR */ }
                                 fgets(buf, sizeof(buf), fp);
-                                strvacat(extdesc, sizeof(extdesc), buf+1, 0);
+                                strvacat(extdesc, sizeof(extdesc), buf+1,
+					 (char *) 0);
                                 i = fgetc(fp);
 			}
                         ungetc(i, fp); /* toss the last one back */
@@ -666,21 +775,29 @@ struct template *template_load(const char *filename)
 				template_lset(t, NULL, "extended_description", extdesc);
 			}
 		}
-		else if (strstr(p, "Description-") == p && t != 0)
+		else if (i18n && strstr(p, "Description-") == p && t != 0)
 		{
-			cp = strstr(p, ".UTF-8: ");
-			if (cp != NULL && cp != p+12)
+			if (strcasestr(p, "Description-C: ") == p)
 			{
-				lang = strndup(p+12, (int) (cp - p - 12));
-				template_lset(t, lang, "description", cp+8);
+				lang = strdup("C");
+				template_lset(t, lang, "description", p+15);
 			}
 			else
 			{
+				cp = strcasestr(p, ".UTF-8: ");
+				if (cp != NULL && cp != p+12)
+				{
+					lang = strndup(p+12, (int) (cp - p - 12));
+					template_lset(t, lang, "description", cp+8);
+				}
+				else
+				{
 #ifndef NODEBUG
-				fprintf(stderr, "Unknown localized field:\n%s\n", p);
+					fprintf(stderr, "Unknown localized field:\n%s\n", p);
 #endif
-				/*  Skip extended description  */
-				lang = NULL;
+					/*  Skip extended description  */
+					lang = NULL;
+				}
 			}
 			extdesc[0] = 0;
 			i = fgetc(fp);
@@ -695,7 +812,8 @@ struct template *template_load(const char *filename)
 			{
 				if (i != ungetc(i, fp)) { /* ERROR */ }
 				fgets(buf, sizeof(buf), fp);
-				strvacat(extdesc, sizeof(extdesc), buf+1, 0);
+				strvacat(extdesc, sizeof(extdesc), buf+1,
+					 (char *) 0);
                                 i = fgetc(fp);
 			}
                         ungetc(i, fp); /* toss the last one back */
